/*
 * audio.c - part of Ivona tools (Ivona synthesizer by Ivo Software)
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Ivona tools is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * Ivona tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ivona tools.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ao/ao.h>
#include "common.h"


static int audio_on;
static ao_device *device;
static int driver;
static ao_sample_format format;
static ao_option *ao_options;

void ao_add_option(char *opt,int is_cli)
{
	char buf[256],*c;
	strcpy(buf,opt);
	c=strchr(buf,':');
	if (!c || !c[1] || c==buf) {
		if (is_cli) {
			fprintf(stderr,"Option [%s]: illegal format\n",buf);
			exit(1);
		}
		return;
	}
	*c++=0;
	ao_append_option(&ao_options,buf,c);
}

int audio_is_live(void)
{
	int driver;
	ao_info *info;
	if (!*ivonaAudioDriver) driver = ao_default_driver_id();
	else driver=ao_driver_id(ivonaAudioDriver);
	if (driver == -1) {
		fprintf(stderr, "Cannot find driver.\n");
		exit(1);
	}
	info=ao_driver_info(driver);
	return (info->type == AO_TYPE_LIVE);
	
}

static void init_audio()
{
	ao_info *info;
	format.bits = 16;
	format.channels = 1;
	format.rate = ivonaRate;
	format.byte_format = AO_FMT_LITTLE;
	if (!*ivonaAudioDriver) driver = ao_default_driver_id();
	else driver=ao_driver_id(ivonaAudioDriver);
	if (driver == -1) {
		fprintf(stderr, "Cannot find driver.\n");
		exit(1);
	}
	info=ao_driver_info(driver);
	if (info->type == AO_TYPE_LIVE) {
		device = ao_open_live(driver, &format, ao_options);
	}
	else {
		device = ao_open_file(driver,ivonaAudioFileName,1,&format,ao_options);
	}
	if (device == NULL) {
		fprintf(stderr, "Error opening audio device.\n");
		exit(1);
	}
}

int ao_callback(char *data,int len)
{
	if (!data) {
		if (audio_on) {
			ao_close(device);
			audio_on=0;
		}
		return 0;
	}
	if (!audio_on) {
		init_audio();
		audio_on=1;
	}
	if (!ao_play(device,data,len)) return 1;
	return 0;
}

void ao_list_drivers(void)
{
        ao_info **info;
        int count,i,j;
        ao_initialize();
        info=ao_driver_info_list(&count);
        for (i=0;i<count;i++) if (/*info[i]->type == AO_TYPE_LIVE &&*/ strcmp(info[i]->short_name,"null")) {
		fprintf(stderr,"%s: %s\n",info[i]->short_name,info[i]->name);
		for (j=0;j<info[i]->option_count;j++) fprintf(stderr,"\t%s\n",info[i]->options[j]);
        }                                                                       
        exit(0);                                                                
}                                                                               
