/*
 * conf.c - part of Ivona tools (Ivona synthesizer by Ivo Software)
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Ivona tools is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * Ivona tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ivona tools.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/stat.h>
#include <ctype.h>
#include <ao/ao.h>

#include "common.h"

static char ivona_args[256];

int ivonaMode=0;

char ivonaServerHost[256]="localhost";
int ivonaServerPort=9123;
char ivonaServerLanguage[4]="pl";
int ivonaServerRate=16000;

char ivonaCliBinary[256];
char ivonaCliVoice[256];
char ivonaCliLanguage[4]="pl";
char *ivonaCliArgs[11];
int ivonaCliRate=16000;

int ivonaWebServicePort=20433;
int ivonaWebServiceAutoStop=1;
int ivonaWebServiceDaemon=0;
int ivonaWebServiceUrlDecode=0;
int ivonaSSIPServerPort=6560;
char ivonaSSIPVoiceName[256]="Pafnucy";
char ivonaSSIPPunctChars[256]="()-";
int ivonaSSIPPunctuation=0;
int ivonaSSIPCapLetRec=2;

char ivonaAudioDriver[256];
char ivonaAudioFileName[256];

int ivonaRate;
char *ivonaLanguage;


#define IVONA_SET(a) if (!strcmp(line,#a)) {strcpy(a,c);return;}
#define IVONA_LANG(a) if (!strcmp(line,#a)) {if (strlen(c)==2)strcpy(a,c);return;}
#define IVONA_BOOL(a,b) if (!strcmp(line,#a)) {a=(*c==b);return;}
#define IVONA_INT(a) if (!strcmp(line,#a)) {a=strtol(c,NULL,10);return;}

static void conf_line(char *line)
{
	char *c,*d;int i;
	
	c=strchr(line,'#');
	if (c) *c=0;
	for (c=d=line;*c;c++) if (!isspace(*c)) d=c+1;
	*d=0;
	while (*line && isspace(*line)) line++;
	if (!*line) return;
	for (c=line;*c && !isspace(*c);c++);
	if (!*c) return;
	*c++=0;
	while (*c && isspace(*c)) c++;
	IVONA_BOOL(ivonaMode,'c');
	IVONA_SET(ivonaServerHost);
	IVONA_SET(ivonaSSIPVoiceName);
	IVONA_INT(ivonaServerPort);
	IVONA_INT(ivonaSSIPServerPort);
	IVONA_LANG(ivonaServerLanguage);
	IVONA_SET(ivonaCliBinary);
	IVONA_SET(ivonaCliVoice);
	IVONA_LANG(ivonaCliLanguage);
	IVONA_INT(ivonaServerRate);
	IVONA_INT(ivonaCliRate);
	
	
	if (!strcmp(line,"ivonaCliArgs")) {
		strcpy(ivona_args,c);
		c=ivona_args;
		for (i=0;*c && i<10;i++) {
			ivonaCliArgs[i]=c;
			for (;*c && !isspace(*c);c++);
			if (*c) {
				*c++=0;
				for (;*c && isspace(*c);c++);
			}
		}
		ivonaCliArgs[i]=NULL;
		return;
	}
	if (!strcmp(line,"ivonaWebServicePort")) {
		if (*c=='o') ivonaWebServicePort=20433;
		else if (*c=='e') ivonaWebServicePort=2222;
		else ivonaWebServicePort=strtol(c,0,10);
		return;
	}
	IVONA_BOOL(ivonaWebServiceAutoStop,'t');
	IVONA_SET(ivonaAudioDriver);
	if (!strcmp(line,"ivonaAudioOption")) {
		ao_add_option(c,0);
		return;
	}
	if (!strcmp(line,"ivonaSSIPPunctuation")) {
		if (!strcmp(c,"none")) ivonaSSIPPunctuation=0;
		else if (!strcmp(c,"some")) ivonaSSIPPunctuation=1;
		else if (!strcmp(c,"all")) ivonaSSIPPunctuation=2;
		return;
	}
	IVONA_SET(ivonaSSIPPunctChars);
	if (!strcmp(line,"ivonaSSIPCapLetRec")) {
		if (!strcmp(c,"none")) ivonaSSIPCapLetRec=0;
		else if (!strcmp(c,"icon")) ivonaSSIPCapLetRec=2;
		else if (!strcmp(c,"spell")) ivonaSSIPCapLetRec=2;
		return;
	}
	
}

static void ivona_ReadConf(void)
{
	char buf[256];
	char *c;
	FILE *f;
	f=fopen("/etc/ivonarc","r");
	if (f) {
		while(fgets(buf,256,f)) conf_line(buf);
		fclose(f);
	}
	c=getenv("HOME");
	if (c) {
		sprintf(buf,"%s/.ivonarc",c);
		f=fopen(buf,"r");
		if (f) {
			while(fgets(buf,256,f)) conf_line(buf);
			fclose(f);
		}
	}
}

static int conf_get_bool(char *str,int fa,int tr,int *err)
{
	if (*str == fa) return 0;
	if (*str == tr) return 1;
	*err=1;
	return 0;
}

static int conf_get_int(char *str,int *err)
{
	int ival;
	char *c;
	ival=strtol(str,&c,10);
	if (!ival || *c) *err=1;
	return ival;
}

static void is_exe(char *fname,char *errstr)
{
	struct stat sb;
	if (!*fname) {
		fprintf(stderr,"%s is not set\n",errstr);
		exit(1);
	}
	if (stat(fname,&sb)) {
		fprintf(stderr,"%s %s not found\n",errstr,fname);
		exit(1);
	}
}

int ivona_LineConf(int argc,char *argv[],int ws)
{
	int i;
	char *lang=NULL;
	int rate=0;
	int xmode=-1;
	char *xargs=NULL;
	int do_help=0;
	
	struct {
		char *name;
		int retval;
		char *help;
	} base_options[]={
		{"autostop",'a',"Web Service autostop mode (true/false)"},
		{"urldecode",'U',"run urldecode in Web Service"},
		{"wport",'W',"Web Service listen port (orca/emacs/integer)"},
		{"daemon",'D',"run Web Service in daemon mode"},
		{"mode",'m',"Ivona mode (server/cli)"},
		{"host",'H',"Ivona server hostname"},
		{"port",'P',"Ivona server port"},
		{"language",'l',"libdumbtts language"},
		{"rate",'r',"Ivona rate",},
		{"binary",'b',"Ivona executable file"},
		{"voice",'v',"Ivona voice file"},
		{"args",'x',"Extra args to Ivona"},
		{"driver",'d',"Audio driver"},
		{"file",'o',"Audio file name (for wav/raw/au drivers)"},
		{"option",'O',"Audio driver option (name:value)"},
		{"list",'L',"Show audio drivers and options"},
		{"help",'h',"Show this help"},
		{NULL,0,NULL}};
	struct option long_options[32];
	char short_options[64],*cs;
	int base_start=0;
	if (ws==2) {
		base_options[2].name="sport";
		base_options[2].retval='S';
		base_options[2].help="SSIP server port (default 6560)";
		base_options[3].help="run SSIP server in daemon mode";
		base_start=2;
	}
	else if (!ws) base_start=4;
	for (cs=short_options,i=base_start;base_options[i].name;i++) {
		*cs++=base_options[i].retval;
		long_options[i].name=base_options[i].name;
		if (strchr("hLDU",base_options[i].retval)) {
			long_options[i].has_arg=0;
		}
		else {
			long_options[i].has_arg=1;
			*cs++=':';
		}
		long_options[i].flag=NULL;
		long_options[i].val=base_options[i].retval;
	}
	*cs=0;
	long_options[i].name=NULL;
	long_options[i].has_arg=0;
	long_options[i].flag=NULL;
	long_options[i].val=0;
	
	ivona_ReadConf();
	
	
	while (!do_help) {
		int c=getopt_long(argc,argv,short_options,long_options,NULL);
		if (c==-1) break;
		if (optarg && strlen(optarg)>255) {
			do_help=1;
			break;
		}
		switch(c) {
			case 'H':
			strcpy(ivonaServerHost,optarg);
			xmode=0;
			break;
			
			case 'P':
			ivonaServerPort=conf_get_int(optarg,&do_help);
			xmode=0;
			break;
			
			case 'm':
			ivonaMode=conf_get_bool(optarg,'s','c',&do_help);
			break;
			
			case 'l':
			if (strlen(optarg)!=2) {
				do_help=1;
				break;
			}
			lang=optarg;
			break;
			
			case 'r':
			rate=conf_get_int(optarg,&do_help);
			break;
			
			case 'b':
			strcpy(ivonaCliBinary,optarg);
			xmode=1;
			break;
			
			case 'v':
			strcpy(ivonaCliVoice,optarg);
			xmode=1;
			break;
			
			case 'x':
			xargs=optarg;
			xmode=1;
			break;
			
			case 'd':
			strcpy(ivonaAudioDriver,optarg);
			break;
			
			case 'O':
			ao_add_option(optarg,1);
			break;
			
			case 'o':
			strcpy(ivonaAudioFileName,optarg);
			break;
			
			case 'h':
			do_help=1;
			break;
			
			case 'L':
			do_help=2;
			break;
			
			case 'W':
			if (*optarg=='o') {
				ivonaWebServicePort=20433;
				ivonaWebServiceAutoStop=1;
			}
			else if (*optarg=='e') {
				ivonaWebServicePort=2222;
				ivonaWebServiceAutoStop=0;
			}
			else ivonaWebServicePort=conf_get_int(optarg,&do_help);
			break;
			
			case 'a':
			ivonaWebServiceAutoStop=conf_get_bool(optarg,'f','t',&do_help);
			break;
			
			case 'D':
			ivonaWebServiceDaemon=1;
			break;
			
			case 'U':
			ivonaWebServiceUrlDecode=1;
			break;
			
			case 'S':
			ivonaSSIPServerPort=conf_get_int(optarg,&do_help);
			break;
			
			default:
			do_help=1;
			break;
		}
	}
	if (do_help==2) {
		ao_list_drivers();
		exit(0);
	}
	if (do_help==1) {
		fprintf(stderr,"Usage: %s [options] [string]\nOptions are:\n",argv[0]);
		for (i=base_start;base_options[i].name;i++) {
			printf("\t-%c --%s\n\t    %s\n",
				base_options[i].retval,base_options[i].name,base_options[i].help);
		}
		exit(0);
	}
	if (xmode>=0) ivonaMode=xmode;
	if (lang) {
		strcpy(ivonaServerLanguage,lang);
		strcpy(ivonaCliLanguage,lang);
	}
	if (rate) {
		ivonaServerRate=rate;
		ivonaCliRate=rate;
	}
	if (xargs) {
		char *c;
		strcpy(ivona_args,xargs);
		c=ivona_args;
		for (i=0;*c && i<10;i++) {
			ivonaCliArgs[i]=c;
			for (;*c && !isspace(*c);c++);
			if (*c) {
				*c++=0;
				for (;*c && isspace(*c);c++);
			}
		}
		ivonaCliArgs[i]=NULL;
	}
	if (ivonaMode) {
		ivonaRate=ivonaCliRate;
		ivonaLanguage=ivonaCliLanguage;
	}
	else {
		ivonaRate=ivonaServerRate;
		ivonaLanguage=ivonaServerLanguage;
		
	}
	
	/* tests */
	if (ivonaMode) {
		is_exe(ivonaCliBinary,"CLI binary");
		is_exe(ivonaCliVoice,"CLI voice");
	}
	else {
		if (!init_addr()) {
			fprintf(stderr,"Bad hostname\n");
			exit(1);
		}
	}
	
	ao_initialize();

	if (ws) {
		if (!audio_is_live()) {
			fprintf(stderr,"File drivers are not allowed\n");
			exit(1);
		}
	}
	else {
		if (!audio_is_live() && !ivonaAudioFileName[0]) {
			fprintf(stderr,"Please set output file name\n");
			exit(1);
		}
	}
	return optind;
}
