/*
 * ivona say.c - part of Ivona tools (Ivona synthesizer by Ivo Software)
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Ivona tools is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * Ivona tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ivona tools.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>

#include <libdumbtts.h>

#include "common.h"

struct dumbtts_conf *dumbconf;

struct speechd_client {
	struct speechd_client *next;
	int sock;
	int settings;
	int status;
	int bufpos,buflen;
	int punctuation;
	int rate;
	int pitch;
	int volume;
	int caplet; /* 0 - none; 1 - spell; 2 - icon */
	int spelling;
	int client_id;
	int notify;
	int prio;
	int msgpos;
	int msgcnt;
	char *name;
	char buffer[8192];
	char msgbuf[65536];
} *clients;

struct speak_message {
	struct speak_message *next;
	int punctuation;
	int rate;
	int pitch;
	int volume;
	int caplet; /* 0 - none; 1 - spell; 2 - icon */
	int spelling;
	int client_id;
	int notify;
	int prio;
	int mode;
	int status; /* 0 - waiting; 1 - canceled; 2 - paused; 3 - postponed */
	int id;
	char message[1];
} *message_queue;

#define MESSAGE_SPEAKING 1
#define MESSAGE_CANCEL 2
#define MESSAGE_PAUSE 3
#define MESSAGE_PAUSED 4


static int client_id_seq;
struct speechd_client *alloc_client(int fd)
{
	struct speechd_client *sd;
	sd=calloc(1,sizeof(*sd));
	sd->next=clients;
	clients=sd;
	sd->sock=fd;
	sd->caplet=1;
	sd->client_id=++client_id_seq;
	return sd;	
}

struct speechd_client **find_client(int fd)
{
	struct speechd_client **ssd;
	for (ssd=&clients;*ssd;ssd=&(*ssd)->next) if ((*ssd)->sock == fd) return ssd;
	return NULL;
}
	
int get_max_fd(int omx)
{
	struct speechd_client *sd;
	for (sd=clients;sd;sd=sd->next) if (sd->sock>omx) omx=sd->sock;
	return omx;
}

void fd_sd_set(fd_set *fs)
{
	struct speechd_client *sd;
	for (sd=clients;sd;sd=sd->next) FD_SET(sd->sock,fs);
	
}

static int message_seq;


static int speak_pid;
static void say_direct(char *str)
{
	speak_pid=fork();
	if (speak_pid < 0) {
		perror("fork");
		exit(1);
	}
	if (!speak_pid) {
		say(str,ao_callback);
		exit(0);
	}
}

void queue_speak_message(struct speechd_client *sd,int mode,char *str)
{
	char buf[256];
	struct speak_message *sm,**ssm;
	int seq=++message_seq;
	

	if (mode) sprintf(buf,"225 OK MESSAGE QUEUED\r\n");
	else sprintf(buf,"225-%d\r\n225 OK MESSAGE_QUEUED\r\n",seq);
	write(sd->sock,buf,strlen(buf));

	/* dla pustych wiadomoci od razu notify i won */

	sm=malloc(sizeof(*sm)+strlen(str));
	sm->punctuation=sd->punctuation;
	sm->rate=sd->rate;
	sm->pitch=sd->pitch;
	sm->volume=sd->volume;
	sm->caplet=sd->caplet;
	sm->spelling=sd->spelling;
	sm->client_id=sd->client_id;
	sm->notify=sd->notify;
	sm->prio=sd->prio;
	sm->mode=mode;
	sm->status=0;
	sm->id=seq;
	strcpy(sm->message,str);
	
	for(ssm=&message_queue;*ssm;ssm=&(*ssm)->next);
	sm->next=NULL;
	*ssm=sm;
}

static int spoken_message;

void notify_client(struct speak_message *sm,int rcode,char *txt)
{
	char buf[256];
	struct speechd_client *sd;
	
	for (sd=clients;sd;sd=sd->next) if (sd->client_id == sm->client_id) break;
	if (!sd || sd->sock == -1) return;
	sprintf(buf,"%d-%d\r\n%d-%d\r\n%d %s\r\n",
		rcode,sm->id,rcode,sm->client_id,rcode,txt);
	write(sd->sock,buf,strlen(buf));
}


void finish_speak()
{

	struct speak_message *sm,**ssm;
	for (ssm=&message_queue;*ssm;ssm=&(*ssm)->next) {
		if ((*ssm)->id == spoken_message) break;
	}
	if (!*ssm) return;
	sm=*ssm;
	if (sm->status == MESSAGE_SPEAKING) { /* speaking, normal break */
		if (sm->notify & 4) notify_client(sm,702,"END");
		*ssm=sm->next;
		free(sm);
		return;
	}
	if (sm->status == MESSAGE_CANCEL) {
		if (sm->notify & 8) notify_client(sm,703,"CANCEL");
		*ssm=sm->next;
		free(sm);
		return;
	}
	if (sm->status == MESSAGE_PAUSE) {
		sm->status = MESSAGE_PAUSED;
		if (sm->notify & 16) notify_client(sm,704,"PAUSE");
		return;
	}
	/* should not be here */
	notify_client(sm,702,"END");
	*ssm=sm->next;
	free(sm);
}

static int get_unichar(char *str,char **ptr)
{
	wchar_t wc;int n;
	wc=*str++ & 255;
	if ((wc & 0xe0)==0xc0) {
		wc &=0x1f;
		n=1;
	}
	else if ((wc & 0xf0)==0xe0) {
		wc &=0x0f;
		n=2;
	}
	else if ((wc & 0xf8)==0xf0) {
		wc &=0x07;
		n=3;
	}
	else if ((wc & 0xfc)==0xf8) {
		wc &=0x03;
		n=4;
	}
	else if ((wc & 0xfe)==0xfc) {
		wc &=0x01;
		n=5;
	}
	else {
		if (ptr) *ptr=str;
		return wc;
	}
	
	while (n--) {
		if ((*str & 0xc0) != 0x80) {
			wc='?';
			break;
		}
		wc=(wc << 6) | ((*str++) & 0x3f);
	}
	if (ptr) *ptr=str;
	return wc;
}

static char *get_dumb_string(struct speak_message *sm)
{
	static int buflen;
	static char *buf;
	if (!buf) {
		buflen=strlen(sm->message)+1024;
		buf=malloc(buflen);
		if (!buf) {
			perror("malloc");
			exit(1);
		}
	}
	if (!sm->mode && sm->spelling) {
		int bufpos=0;
		char *s=sm->message;
		wchar_t wc;
		int blank=0;
		while(*s) {
			wc=get_unichar(s,&s);
			
			if (wc<0x80 && isspace(wc)) {
				blank=1;
				continue;
			}
			if (bufpos>=buflen-256) {
				buflen+=1024;
				buf=realloc(buf,buflen);
				if (!buf) {
					perror("realloc");
					exit(1);
				}
			}
			if (bufpos) {
				buf[bufpos++]=',';
				buf[bufpos++]=' ';
				buf[bufpos]=0;
			}
			if (blank) {
				dumbtts_WCharString(dumbconf,0x20,buf+bufpos,buflen-bufpos,sm->caplet,NULL);
				while(buf[bufpos]) bufpos++;
				buf[bufpos++]=',';
				buf[bufpos++]=' ';
				buf[bufpos]=0;
			}
			dumbtts_WCharString(dumbconf,wc,buf+bufpos,buflen-bufpos,sm->caplet,NULL);
			while(buf[bufpos]) bufpos++;
			blank=0;
		}
		return buf;
	}
	
	
	for (;;) {
		int len;
		switch(sm->mode) {
			case 1: /* char */
			len=dumbtts_CharString(dumbconf,sm->message,buf,buflen,sm->caplet,NULL);
			break;
			case 2: /* key */
			len=dumbtts_KeyString(dumbconf,sm->message,buf,buflen,sm->caplet,NULL);
			break;
			default:
			len=dumbtts_GetString(dumbconf,sm->message,buf,buflen,sm->punctuation,ivonaSSIPPunctChars,",.:;?!");
		}
		if (!len) break;
		if (len<0) return "error";
		buflen=len+1024;
		buf=realloc(buf,buflen);
		if (!buf) {
			perror("realloc");
			exit(0);
		}
	}
	return buf;
}

void dequeue_message()
{
	struct speechd_client *sd;
	char buf[256];
	struct speak_message *sm,**ssm;
	
	for (ssm=&message_queue;*ssm;) {
		if ((*ssm)->status == MESSAGE_CANCEL) {
			sm=*ssm;
			*ssm=sm->next;
			if (sm->notify & 8) notify_client(sm,703,"CANCEL");
			free(sm);
		}
		else ssm=&(*ssm)->next;
	}
			
	
	sm=message_queue;
	if (!sm) return;
	//message_queue=sm->next;
	for (sd=clients;sd;sd=sd->next) if (sd->client_id == sm->client_id) break;
	say_direct(get_dumb_string(sm));
	if (sd && sd->sock>=0) {
		if (sm->notify & 2) {
			sprintf(buf,"701-%d\r\n701-%d\r\n701 BEGIN\r\n",
				sm->id,sm->client_id);
			write(sd->sock,buf,strlen(buf));
		}
	}
	spoken_message=sm->id;
	sm->status=MESSAGE_SPEAKING;
	
}

void cancel_all_messages(void)
{
	struct speak_message *sm;
	for (sm=message_queue;sm;sm=sm->next) {
		if (!sm->status && (sm->notify & 2)) notify_client(sm,701,"BEGIN");
		sm->status=MESSAGE_CANCEL;
	}
	if (speak_pid) kill(speak_pid,9);
	
}

#define sets(a) (!strcasecmp(args[2],(a)))

int get_command(struct speechd_client *sd)
{
	char *c;
	int got=0;
	char *args[8];
	char buf[256];
	int nargs;
	
	int proper_int(int a)
	{
		int nr;char *c;
		if (nargs != a) return 0;
		nr=strtol(args[3],&c,10);
		if (nr< -100 || nr > 100 || *c) return 0;
		return 1;
	}
	
	
	for (c=sd->buffer;c<sd->buffer+sd->buflen;c++) if (*c=='\n') {
		*c++=0;
		got=c-sd->buffer;
		break;
	}
	if (!got) return 0;
	for (nargs=0,c=sd->buffer;nargs<8;nargs++) {
		while (*c && isspace(*c)) c++;
		if (!*c) break;
		args[nargs]=c;
		while (*c && !isspace(*c)) c++;
		if (*c) *c++=0;
	}
	
	if (!nargs) {
bad_cmd:	sprintf(buf,"500 BAD COMMAND\r\n");
write_rsp:	write(sd->sock,buf,strlen(buf));
		goto dalej;
	}
	if (!strcasecmp(args[0],"quit")) {
		sprintf(buf,"240 OK GOOD BYE\r\n");
		write(sd->sock,buf,strlen(buf));
		close(sd->sock);
		sd->sock=-1;
		return 1;
	}
	if (!strcasecmp(args[0],"set")) {
		
		if (nargs < 4) goto bad_cmd;
		if (sets("client_name")) {
			if (nargs != 4) goto bad_cmd;
			if (sd->name) sprintf(buf,"500 NAME ALREADY SET\r\n");
			else {
				if (strlen(args[3])>128) {
					sprintf(buf,"500 NAME TOO LONG\r\n");
				}
				else {
					sd->name=strdup(args[3]);
					sprintf(buf,"200 CLIENT NAME SET %s\n",sd->name);
				}
			}
			goto write_rsp;
		}
		if (sets("language")) {
			/* ignore */
			sprintf(buf,"201 OK LANGUAGE SET\r\n");
			goto write_rsp;
		}
		if (sets("output_module")) {
			/* ignore */
			sprintf(buf,"201 OK OUTPUT MODULE SET\r\n");
			goto write_rsp;
		}
		if (sets("voice") || !strcasecmp(args[2],"synthesis_voice")) {
			/* ignore */
			sprintf(buf,"209 OK VOICE SET\r\n");
			goto write_rsp;
		}
		if (sets("pitch")) {
			if (!proper_int(4)) goto badparam;
			sd->pitch=atoi(args[3]);
			sprintf(buf,"204 OK PITCH SET\r\n");
			goto write_rsp;
		}
		if (sets("rate")) {
			if (!proper_int(4)) goto badparam;
			sd->rate=atoi(args[3]);
			sprintf(buf,"203 OK RATE SET\r\n");
			goto write_rsp;
		}
		if (sets("volume")) {
			if (!proper_int(4)) goto badparam;
			sd->volume=atoi(args[3]);
			sprintf(buf,"218 OK VOLUME SET\r\n");
			goto write_rsp;
		}
		if (sets("punctuation")) {
			if (nargs != 4) goto badparam;
			if (!strcasecmp(args[3],"none")) sd->punctuation=0;
			else if (!strcasecmp(args[3],"some")) sd->punctuation=1;
			else if (!strcasecmp(args[3],"all")) sd->punctuation=2;
			else goto badparam;
			sprintf(buf,"205 OK PUNCTUATION SET\r\n");
			goto write_rsp;
		}
		if (sets("cap_let_recogn")) {
			if (nargs != 4) goto badparam;
			if (!strcasecmp(args[3],"none")) sd->caplet=0;
			else if (!strcasecmp(args[3],"icon")) sd->caplet=2;
			else if (!strcasecmp(args[3],"spell")) sd->caplet=2;
			else goto badparam;
			sprintf(buf,"206 OK CAP LET RECOGNITION SET\r\n");
			goto write_rsp;
		}
		if (sets("spelling")) {
			if (nargs != 4) goto badparam;
			if (!strcasecmp(args[3],"off")) sd->spelling=0;
			else if (!strcasecmp(args[3],"on")) sd->spelling=1;
			else goto badparam;
			sprintf(buf,"207 OK SPELLING SET\r\n");
			goto write_rsp;
		}
		if (sets("pause_context")) {
			/* ignore */
			sprintf(buf,"217 OK PAUSE CONTEXT SET\r\n");
			goto write_rsp;
		}
		if (sets("notification")) {
			int mask=0;
			if (nargs != 5) goto badparam;
			if (!strcasecmp(args[3],"index_mark")) mask=1;
			else if (!strcasecmp(args[3],"begin")) mask=2;
			else if (!strcasecmp(args[3],"end")) mask=4;
			else if (!strcasecmp(args[3],"cancel")) mask=8;
			else if (!strcasecmp(args[3],"pause")) mask=16;
			else if (!strcasecmp(args[3],"resume")) mask=32;
			else if (!strcasecmp(args[3],"all")) mask=63;
			else goto badparam;
			if (!strcasecmp(args[4],"on")) {
				sd->notify = sd->notify | mask;
			}
			else if (!strcasecmp(args[4],"off")) {
				sd->notify = sd ->notify & ~mask;
			}
			else goto badparam;
			sprintf(buf,"220 OK NOTIFICATION SET\r\n");
			goto write_rsp;
		}
badparam:	sprintf(buf,"500 UNKNOWN PARAMETER\r\n");
		goto write_rsp;
	}
	if (!strcasecmp(args[0],"list")) {
		if (nargs != 2) goto bad_cmd;
		if (!strcasecmp(args[1],"output_modules")) {
			sprintf(buf,"250-ivona\r\n250 OK MODULE LIST SENT\r\n");
			goto write_rsp;
		}
		if (!strcasecmp(args[1],"voices")) {
			sprintf(buf,"249-MALE1\r\n249 OK VOICE LIST SENT\r\n");
			goto write_rsp;
		}
		if (!strcasecmp(args[1],"SYNTHESIS_VOICES")) {
			sprintf(buf,"249-%s %s none\r\n249 OK VOICE LIST SENT\r\n",ivonaSSIPVoiceName,ivonaLanguage);
			goto write_rsp;
		}
		goto badparam;
	}
	if (!strcasecmp(args[0],"char")) {
		if (nargs != 2) goto badparam;
		queue_speak_message(sd,1,args[1]);
		goto dalej;
	}
	if (!strcasecmp(args[0],"key")) {
		if (nargs != 2) goto badparam;
		queue_speak_message(sd,2,args[1]);
		goto dalej;
	}
	if (!strcasecmp(args[0],"speak")) {
		sd->status=1;
		sd->msgpos=sd->msgcnt=0;
		sprintf(buf,"230 OK RECEIVING DATA\r\n");
		goto write_rsp;
	}
	if (!strcasecmp(args[0],"cancel") || !strcasecmp(args[0],"stop")) {
		/* parameter ignored */
		cancel_all_messages();
		sprintf(buf,"214 OK CANCELLED\r\n");
		goto write_rsp;
	}
	if (!strcasecmp(args[0],"help")) {
		strcpy(buf,"200-SET\r\n200-KEY\r\n200-CHAR\r\n200-SPEAK\r\n200-CANCEL\r\n200-STOP\r\n200-LIST\r\n200-QUIT\r\n200-HELP\r\n200 OK\r\n");
		goto write_rsp;
	}	
	goto bad_cmd;
dalej:	
	if (got < sd->buflen) {
		memcpy(sd->buffer,sd->buffer+got,sd->buflen-got);
	}
	sd->buflen-=got;
	if (sd->buflen < 0) sd->buflen=0;
	sd->bufpos=sd->buflen;
	return 1;
}

#undef sets

void remove_rdot(char *c)
{
	char *d;int znak,blank;
	d=c;
	for (;*c;c++) if (!isspace(*c)) break;
	blank=0;
	for (;*c;) {
		znak=*c++;
		if (isspace(znak)) {
			blank=1;
		}
		else {
			if (blank) *d++=' ';
			*d++=znak;
			blank=0;
		}
		if (znak == '\n' && *c=='.' && c[1]=='.') c++;
	}
	*d=0;
}

void get_message(struct speechd_client *sd)
{
	char *c;
	int eon=0,left;
	
	for (c=sd->buffer;c<sd->buffer+sd->buflen;c++) {
		sd->msgbuf[sd->msgpos++]=*c;
		if (*c=='\n') {
			if (sd->msgbuf[sd->msgcnt]=='.' && sd->msgcnt+1 < sd->msgpos && (sd->msgbuf[sd->msgcnt+1]=='\r' || sd->msgbuf[sd->msgcnt+1]=='\n')) {
				sd->msgbuf[sd->msgcnt]=0;
				c++;
				eon=1;
				break;
			}
			sd->msgcnt=sd->msgpos;
		}
	}
	sd->msgbuf[sd->msgpos]=0;
	left=sd->buflen-(c-sd->buffer);
	if (left > 0) {
		memcpy(sd->buffer,c,left);
	}
	if (left<0) left=0;
	sd->bufpos=sd->buflen=left;
	if (!eon) return;
	remove_rdot(sd->msgbuf);
	queue_speak_message(sd,0,sd->msgbuf);
	sd->msgpos=sd->msgcnt=0;
	sd->status=0;
}


void do_client(struct speechd_client *sd)
{
	int n;
	n=read(sd->sock,sd->buffer+sd->bufpos,8192-sd->buflen);
	if (n<=0) {
		close(sd->sock);
		sd->sock=-1;
		return;
	}
	sd->buflen+=n;
	sd->bufpos=sd->buflen;
	for(;;) {
		if (sd->buflen<=0) return;
		if (!sd->status) {
			if (!get_command(sd)) break;
			if (sd->sock == -1) return; /* closed */
			continue;
		}
		if (sd->status == 1) {
			/* getting message */
			get_message(sd);
			if (sd->sock == -1) return; /* closed */
			continue;
		}
		sd->bufpos=sd->buflen=0;
		sd->status=0;			
	}
}


void free_client(struct speechd_client *sd)
{
	if (sd->name) free(sd->name);
	free(sd);
}


void do_clients(fd_set *fs)
{
	struct speechd_client *sd,**ssd;
	for (sd=clients;sd;sd=sd->next) {
		if (FD_ISSET(sd->sock,fs)) {
			do_client(sd);
		}
	}
	for (ssd=&clients;*ssd;) {
		if ((*ssd)->sock != -1) {
			ssd=&(*ssd)->next;
			continue;
		}
		sd=*ssd;
		*ssd=sd->next;
		free_client(sd);
	}
}

int main(int argc,char *argv[])
{
	int gniazdko;
	struct sockaddr_in sina;
	int reuse=-1;

	ivona_LineConf(argc,argv,2);
	if (ivonaWebServiceDaemon) {
		close(0);
		close(1);
		close(2);
		if (fork()) exit(0);
	}

	gniazdko=socket(PF_INET,SOCK_STREAM,0);
	if (gniazdko<0) {
		perror("socket");
		exit(0);
	}
	bzero(&sina,sizeof(sina));
	sina.sin_family = AF_INET;
	sina.sin_port = htons(ivonaSSIPServerPort);
	fcntl(gniazdko,F_SETFL,O_NONBLOCK);
	setsockopt(gniazdko,SOL_SOCKET,SO_REUSEADDR,(void *)&reuse,sizeof(reuse));
	if (bind(gniazdko, (struct sockaddr *)&sina,sizeof(sina))) {
		perror("bind");
		close(gniazdko);
		exit(0);
	}
	if (listen(gniazdko,128)) {
		perror("listen");
		close(gniazdko);
		exit(0);
	}
	dumbconf=dumbtts_TTSInit("pl");
	for(;;) {
		fd_set rdfs;
		int retval;
		struct timeval tv;
		int fd;
		if (speak_pid) {
			retval=waitpid(speak_pid,NULL,WNOHANG);
			if (retval == speak_pid) {
				finish_speak();
				speak_pid=0;
			}
		}
		if (!speak_pid) dequeue_message();
		
		FD_ZERO(&rdfs);
		FD_SET(gniazdko,&rdfs);
		fd_sd_set(&rdfs);
		tv.tv_sec = 0;
		tv.tv_usec=100000;
		retval=select(get_max_fd(gniazdko)+1,&rdfs,NULL,NULL,&tv);
		fd=accept(gniazdko,0,0);
		if (fd>=0) {
			alloc_client(fd);
			listen(gniazdko,128);
		}
		do_clients(&rdfs);
	}
	return 0;
}
