/*
 * speak.c - part of Ivona tools (Ivona synthesizer by Ivo Software)
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Ivona tools is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * Ivona tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ivona tools.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>

#include "common.h"

static struct sockaddr_in sinadr;

int init_addr(void)
{
	if (!inet_aton(ivonaServerHost,&sinadr.sin_addr)) {
		struct hostent *h = gethostbyname(ivonaServerHost);
		if (!h) return 0;
		memcpy(&sinadr.sin_addr, h->h_addr, sizeof(struct in_addr));
		endhostent();
	}
	sinadr.sin_family = AF_INET;
	sinadr.sin_port = htons(ivonaServerPort);
	
	return 1;
}

static void say_server(char *str,int (*callback)(char *,int))
{
	int fd,len,n;
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) return;
	if (connect(fd, (struct sockaddr *) &sinadr, sizeof(sinadr)) < 0) {
		close(fd);
		return;
	}
	write(fd,str,len=strlen(str));
	if (!len || str[len-1] != '\n') write(fd,"\n",1);
	for (;;) {
		char wave_buffer[8192];
		n=read(fd,wave_buffer,8192);
		if (n<=0) break;
		if (callback(wave_buffer,n)) {
			break;
		}
	}
	close(fd);
	callback(NULL,0);
}

static void say_standalone(char *str,int (*callback)(char *,int))
{
	int pid,n;
	
	int pipe_out[2];
	int pipe_in[2];
	
	if (pipe(pipe_out)) {
		perror("pipe");
		exit(1);
	}
	if (pipe(pipe_in)) {
		perror("pipe");
		exit(1);
	}
	pid=fork();
	if (pid<0) {
		perror("fork");
		exit(1);
	}
	if (!pid) {
		char *args[32];
		int i,n;
		char *c;
		close(pipe_out[0]);
		close(pipe_in[1]);
		dup2(pipe_out[1],1);
		dup2(pipe_in[0],0);
		
		c=strrchr(ivonaCliBinary,'/');
		if (c) c++;else c=ivonaCliBinary;
		n=0;
		args[n++]=c;
		args[n++]="-l";
		args[n++]=ivonaCliVoice;
		for (i=0;ivonaCliArgs[i];i++) args[n++]=ivonaCliArgs[i];
		args[n++]="-f";
		args[n++]="-";
		args[n++]="-";
		args[n++]=NULL;
		execv(ivonaCliBinary,args);
		exit(1);
	}
	close(pipe_out[1]);
	close(pipe_in[0]);
	write(pipe_in[1],str,strlen(str));
	close(pipe_in[1]);
	for (;;) {
		char wave_buffer[8192];
		n=read(pipe_out[0],wave_buffer,8192);
		if (n<=0) break;
		if (callback(wave_buffer,n)) {
			kill(pid,9);
			break;
		}
	}
	close(pipe_out[0]);
	callback(NULL,0);
	waitpid(pid,NULL,0);

}

void say(char *str,int (*callback)(char *,int))
{
	if (ivonaMode) say_standalone(str,callback);
	else say_server(str,callback);
}
